--------------------------------------------------------------------------
-- CharDump
-- Dump character statistics to text file
-- V0.42 to 0.47 - no real changes, just API version
-- V0.41 - 6.5.2016 (CP instead of V level)
-- V0.4  - 27.2.2016 (added alias)
-- V0.32 - 9.10.2015 (no change from 0.3)
-- V0.3  - 11.8.2015 (no change from 0.2)
-- V0.2  - 3.3.2015
-- V0.1  - 1.3.2015
--
-- While the code is my own I looked at a couple of other apps to 
-- get an idea of how things could be done, for example
-- Clock, Craft Research Timer, Combat Log Statistics, Loot Summary
-- I also looked at the first app example from http://wiki.esoui.com
--

------------------------------------------------------------

CharDump = {
  Name        = "CharDump",
  Version     = "0.47",
  SaveFileVersion = 2,
  TextBuffer  = {},
  MaxLines    = 1000,
  Initialized = false,
  LabelHeight  = 20,
  LabelWidth   = 0,  -- will be changed later
  SliderHeight = 0,  -- will be changed later  
  SliderWidth  = 20,  
  TextWindowWidth  = 0,  -- will be changed later
  TextWindowHeight = 0,  -- will be changed later  
}
------------------------------------------------------------

local local_default_settings = {
  Output = {  }
}

local global_default_settings = {
  WindowX = 100,
  WindowY = 100,
  WindowWidth  = 400,
  WindowHeight = 400,
  ShowChat  = true,
  ShowText  = true,
  WriteFile = true,
  IsLocked  = false,
  AllSkills = false,
  MaxHistoryLines = 2000
}


------------------------------------------------------------
-- I did not find a function to convert the SKILLType code to a string
-- If there is one, this function is obsolete

local function GetSkillTypeString(skilltypeid)
  -- is there a case stement in the lua language?
  skillstringout = "ERROR"
  if (skilltypeid == SKILL_TYPE_NONE) then
    skillstringout = "NONE"
  else
    if (skilltypeid == SKILL_TYPE_CLASS) then
      skillstringout = "Class"
	else
      if (skilltypeid == SKILL_TYPE_WEAPON) then
        skillstringout = "Weapon"
	  else
        if (skilltypeid == SKILL_TYPE_ARMOR) then
          skillstringout = "Armor"
	    else
          if (skilltypeid == SKILL_TYPE_WORLD) then
            skillstringout = "World"
	      else
            if (skilltypeid == SKILL_TYPE_GUILD) then
              skillstringout = "Guild"
	        else
              if (skilltypeid == SKILL_TYPE_AVA) then
                skillstringout = "Alliance"
	          else
                if (skilltypeid == SKILL_TYPE_RACIAL) then
                  skillstringout = "Racial"
	            else
                  if (skilltypeid == SKILL_TYPE_TRADESKILL) then
                    skillstringout = "Crafting"
	              else
                    if (skilltypeid == SKILL_TYPE_CHAMPION) then
                      skillstringout = "Champion"
	                else
                      skillstringout = "Unknown"
	                end
	              end
	            end
	          end
	        end
	      end
	    end
	  end
	end
  end
  return skillstringout
end

-------------------------------------------

local function SaveWindowPosition(self)
  CharDump.GlobalSettings.WindowX = math.floor(self:GetLeft())
  CharDump.GlobalSettings.WindowY = math.floor(self:GetTop())
end

local function ScrollToEnd()
  CharDump_TextW:SetScrollPosition(0)
end

local function CreateWindow()
  CharDump.LabelWidth       = CharDump.GlobalSettings.WindowWidth
  CharDump.SliderHeight     = CharDump.GlobalSettings.WindowHeight - CharDump.LabelHeight
  --CharDump.TextWindowWidth  = CharDump.GlobalSettings.WindowWidth - CharDump.SliderWidth
  CharDump.TextWindowWidth  = CharDump.GlobalSettings.WindowWidth -- we dont use the slider atm
  CharDump.TextWindowHeight = CharDump.SliderHeight

  -- top level window - this window conatons all other windows
  CharDump_TLW = WINDOW_MANAGER:CreateTopLevelWindow("CharDump_TLW")
  CharDump_TLW:SetMouseEnabled(true)
  CharDump_TLW:SetHidden(not CharDump.GlobalSettings.ShowText)
  CharDump_TLW:SetClampedToScreen(false)
  CharDump_TLW:SetMovable(not CharDump.GlobalSettings.IsLocked)
  CharDump_TLW:SetDimensions(CharDump.GlobalSettings.WindowWidth,CharDump.GlobalSettings.WindowHeight)
  CharDump_TLW:SetAnchor(TOPLEFT, GuiRoot, TOPLEFT, CharDump.GlobalSettings.WindowX, CharDump.GlobalSettings.WindowY)
  CharDump_TLW:SetHandler("OnMouseUp",     function(self) SaveWindowPosition(self) end)
  CharDump_TLW:SetHandler("OnMouseWheel" , function(self,delta)
		                                     CharDump_TextW:SetScrollPosition(CharDump_TextW:GetScrollPosition() + delta)
	                                       end)
 
  -- label for title
  CharDump_LabelW = WINDOW_MANAGER:CreateControl("CharDump_Label",CharDump_TLW,CT_LABEL)
  CharDump_LabelW:SetHidden(false)
  CharDump_LabelW:SetFont("ZoFontGame")	
  CharDump_LabelW:SetColor(0,255,255,1.5)
  CharDump_LabelW:SetDimensions(CharDump.LabelWidth,CharDump.LabelHeight)
  CharDump_LabelW:SetAnchor(TOPLEFT,CharDump_TLW,TOPLEFT,0,0)
  CharDump_LabelW:SetText("CharDump output window")

  -- background for title
  CharDump_LabelBGW = WINDOW_MANAGER:CreateControl("CharDump_LabelBackground",CharDump_TLW,CT_BACKDROP)
  CharDump_LabelBGW:SetDimensions(CharDump.LabelWidth,CharDump.LabelHeight)
  CharDump_LabelBGW:SetCenterColor(0.1,0.1,0.1,1)
  CharDump_LabelBGW:SetEdgeColor(0,0,0,1)
  CharDump_LabelBGW:SetEdgeTexture("", 8, 1, 1)
  CharDump_LabelBGW:SetAlpha(0.8)
  CharDump_LabelBGW:SetAnchor(TOPLEFT,CharDump_TLW,TOPLEFT,0,0)
  
  -- scroll window
  CharDump_ScrollW = WINDOW_MANAGER:CreateControl("CharDump_Scroll", CharDump_TLW, CT_SCROLL)
  CharDump_ScrollW:SetDimensions(CharDump.TextWindowWidth,CharDump.TextWindowHeight)
  CharDump_ScrollW:SetAnchor(TOPLEFT,CharDump_TLW,TOPLEFT,0,CharDump.LabelHeight)
  CharDump_ScrollW:SetHorizontalScroll(0)
 
  -- scroll slider
  --CharDump_ScrollSliderW = WINDOW_MANAGER:CreateControl("CharDump_ScrollSlider", CharDump_TLW, CT_SLIDER)
  --CharDump_ScrollSliderW:SetDimensions(CharDump.SliderWidth,CharDump.SliderHeight)
  --CharDump_ScrollSliderW:SetAnchor(RIGHT,CharDump_TLW,RIGHT,0,0)
  --CharDump_ScrollSliderW:SetMouseEnabled(true)
  --CharDump_ScrollSliderW:SetThumbTexture(tex,tex,tex,30,10,1,0,1,0)
  --CharDump_ScrollSliderW:SetValue(0)
  --CharDump_ScrollSliderW:SetValueStep(2)
--  CharDump_ScrollSliderW:SetMinMax(0,CL:GetHeight())
  --CharDump_ScrollSliderW:SetHandler("OnValueChanged",function(self,value,eventReason)
  --		CharDump_ScrollW:SetVerticalScroll(value*10)
  --		CharDump.ScrollOffset = value			
  --	end)
		
		
  -- text window
  --CharDump_TextW = WINDOW_MANAGER:CreateControlFromVirtual("CharDump_Text",CharDump_ScrollW,"ZO_DefaultEditMultiLineForBackdrop")
  CharDump_TextW = WINDOW_MANAGER:CreateControl("CharDump_Text",CharDump_ScrollW,CT_TEXTBUFFER)
  CharDump_TextW:SetHidden(false)
  CharDump_TextW:SetFont("ZoFontGame")
  CharDump_TextW:SetDimensions(CharDump.TextWindowWidth,CharDump.TextWindowHeight)
  CharDump_TextW:SetAnchor(TOPLEFT,CharDump_ScrollW,TOPLEFT,0,0)

  CharDump_TextW:SetLinkEnabled(true)
  CharDump_TextW:SetMaxHistoryLines(CharDump.GlobalSettings.MaxHistoryLines)
  --CharDump_TextW:SetCopyEnabled(true)
  --CharDump_TextW:SetMaxInputChars(5000)
  
  --CharDump_TextW:SetText("Line 1")
  --CharDump_TextW:AddMessage("Line 1")

  
  -- background for text and slider
  CharDump_BGW = WINDOW_MANAGER:CreateControl("CharDump_TextBackgroud",CharDump_TLW,CT_BACKDROP)
  CharDump_BGW:SetDimensions(CharDump.TextWindowWidth,CharDump.TextWindowHeight)
  CharDump_BGW:SetCenterColor(0.1,0.1,0.1,1)
  CharDump_BGW:SetEdgeColor(0,0,0,1)
  CharDump_BGW:SetEdgeTexture("", 8, 1, 1)
  CharDump_BGW:SetAlpha(0.3)
  CharDump_BGW:SetAnchor(TOPLEFT,CharDump_TLW,TOPLEFT,0,CharDump.LabelHeight)
end

------------------------------------------------------------------

local function LockWindow()    CharDump.GlobalSettings.IsLocked   = true  end
local function UnlockWindow()  CharDump.GlobalSettings.IsLocked   = false end
local function ShowChat()      CharDump.GlobalSettings.ShowChat   = true  end
local function HideChat()      CharDump.GlobalSettings.ShowChat   = false end
local function ShowText()      CharDump.GlobalSettings.ShowText   = true
                               CharDump_TLW:SetHidden(false)  end
local function HideText()      CharDump.GlobalSettings.ShowText   = false 
                               CharDump_TLW:SetHidden(true)   end
local function WriteFile()     CharDump.GlobalSettings.WriteFile  = true  end
local function NoFile()        CharDump.GlobalSettings.WriteFile  = false end
local function AllSkills()     CharDump.GlobalSettings.AllSkills  = true  end
local function LearnedSkills() CharDump.GlobalSettings.AllSkills  = false end

------------------------------------------------------------

local function OutputData()
  -- this causes the data to be written to the save file
  if (CharDump.GlobalSettings.WriteFile) then
    --CharDump.Settings = CharDump.TmpSettings
    CharDump.Settings.Output = {}
	
    for line = 0, #CharDump.TmpSettings.Output+1  do
      CharDump.Settings.Output[line] = CharDump.TmpSettings.Output[line]
	end
  end
  
  -- output to chat
  if (CharDump.GlobalSettings.ShowChat) then
    for line = 0, #CharDump.TmpSettings.Output+1  do
      CHAT_SYSTEM:AddMessage(CharDump.TmpSettings.Output[line])
    end	  
  end
  
  -- output to window  
  if (CharDump.GlobalSettings.ShowText) then
    ScrollToEnd()
    for line = 0, #CharDump.TmpSettings.Output+1  do
      CharDump_TextW:AddMessage(CharDump.TmpSettings.Output[line])  
	end
  end
  
  -- copy to clipboard
--  if (CharDump.GlobalSettings.ShowText) then
--  local mystring = ""
	
--  for line = 0, #CharDump.Settings.Header+1 do
--    if (CharDump.Settings.Header[line] ~= nil) then
--      mystring = mystring .. CharDump.Settings.Header[line] .. "\n\r"
--	end
--  end	  
--  for line = 0, #CharDump.Settings.Output+1  do
--    if (CharDump.Settings.Output[line] ~= nil) then
--      mystring = mystring .. CharDump.Settings.Output[line] .. "\n\r"
--	end
 -- end

--  CharDump_TextW:SetText(mystring)
--end
  
end

-----------------------------------------------------------------
-- This is where we do the actual work: dump the character sheet

local function AddLine(outstring, myline)
  CharDump.TmpSettings.Output[myline] = outstring
  myline = myline+1
  return myline
end

local function DumpCharacter()
  --------------------------------------
  -- create correct data structure
  
  CharDump.TmpSettings = local_default_settings
  CharDump.TmpSettings.Output = {}
	
  --------------------------------------
  -- Header
  
  -- name, class and level 
  local currentline=0
  
  local charactername  = GetUnitName("player")
  local characterclass = GetUnitClass("player")
  local characterlevel = GetUnitLevel("player")
  local championpoints = GetUnitVeteranRank("player") 
 
  if (championpoints == 0) then
	levelstring = "level " .. characterlevel
  else
    levelstring = championpoints .. " CP" 
  end
  currentline = AddLine("Character sheet for " .. charactername .. " (" .. levelstring .. " " .. characterclass .. ")", currentline)

  
  -- write stats
  local health  = GetAttributeSpentPoints(1) 
  local magicka = GetAttributeSpentPoints(2) 
  local stamina = GetAttributeSpentPoints(3) 
  currentline = AddLine(GetDateStringFromTimestamp(GetTimeStamp()) .. " " .. GetTimeString(), currentline)
  currentline = AddLine("Attribute points spent:  ", currentline)
  currentline = AddLine("Health:   " .. health,  currentline)
  currentline = AddLine("Magicka:  " .. magicka, currentline)
  currentline = AddLine("Stamina:  " .. stamina, currentline)
  
  -- @FIXME write larger header with general character information like level, equipment attributes etc...
  
  local nskilltypes = GetNumSkillTypes() 
  for skilltype = 1, nskilltypes do 	  	  
	  local skillstring = GetSkillTypeString(skilltype)  
      currentline = AddLine(skillstring .. " skill lines", currentline)      
	  currentline = AddLine("====================", currentline)
      
      local nskilllines = GetNumSkillLines(skilltype) 
      for skillline = 1, nskilllines do 
	    local skillname, skillrank = GetSkillLineInfo(skilltype, skillline)

		currentline = AddLine(skillname .. " rank " .. skillrank, currentline)
		currentline = AddLine("--------------------", currentline)
		
        local nabilities  = GetNumSkillAbilities(skilltype, skillline)	
        local gotultimate = 0		
        for ability = 1, nabilities do
		  local abilityname, abilitytexture, abilityrank, ispassive, isultimate, ispurchased, progression = GetSkillAbilityInfo(skilltype, skillline, ability)
		 
          local passivestring = " (passive)"	
		  local basename, morph, rank = GetAbilityProgressionInfo(progression)
		 
          if (isultimate) then
		    gotultimate = 1
	      end
		 
		  if (ispassive) then
		    -- passive skill
			
   		    local currentupgradelevel, maxupgradelevel = GetSkillAbilityUpgradeInfo(skilltype, skillline, ability)
		    if (currentupgradelevel == nil) then
		      upgradestring = ""
		    else
		      upgradestring = " at rank " .. currentupgradelevel 
		    end
			
		    if (ispurchased) then
              currentline = AddLine(abilityname .. passivestring .. upgradestring, currentline)
			else
			  if (CharDump.GlobalSettings.AllSkills) then
                currentline = AddLine(abilityname .. passivestring .. ", not learned", currentline)
		      end
			end
	      else
		    -- active skill
			if (isultimate) then
 			  skilllinerankstring = "(U) "
			else
 			  skilllinerankstring = "(" .. ability-gotultimate .. ") "
			end
				
		    if (ispurchased) then
			  progressionname, progressionmorph, progressionrank = GetAbilityProgressionInfo(progression)

    		  if (basename == abilityname) then  
                currentline = AddLine(skilllinerankstring .. abilityname .. "(at rank ".. progressionrank ..")", currentline)
			  else
                currentline = AddLine(skilllinerankstring .. abilityname .. " (morph of " .. basename .. ", rank " .. progressionrank .. ")", currentline)
			  end
			else
			  if (CharDump.GlobalSettings.AllSkills) then
                currentline = AddLine(skilllinerankstring .. abilityname .. " (not learned)", currentline)
			  end
			end
		  end
		end
		currentline = AddLine("====================", currentline)
	  end
	  
	  
	  -- list champion skills
	  if (skilltype == SKILL_TYPE_CHAMPION) then
	    local nconstellations = GetNumChampionDisciplines()
        for constellation = 1, nconstellations do
		  local constellationname   = GetChampionDisciplineName(constellation)
		  local constellationpoints = GetNumPointsSpentInChampionDiscipline(constellation)
		  currentline = AddLine("====================", currentline)		  
          currentline = AddLine("Constellation: " .. constellationname .. " (" .. constellationpoints .. " CP spent)", currentline)
		  currentline = AddLine("--------------------", currentline)

		  local nchampionskills = GetNumChampionDisciplineSkills(constellation)
          for championskill = 1, nchampionskills do
		    local championskillname   = GetChampionSkillName(constellation, championskill)
			local championskillpoints = GetNumPointsSpentOnChampionSkill(constellation, championskill)
			local unlocklevel = GetChampionSkillUnlockLevel(constellation, championskill)
			
			local passivestring = ""
			if unlocklevel == nil then
			  -- not a passive
			  if (championskillpoints > 0) then
			    currentline = AddLine(championskillname .. ": " .. championskillpoints .. " points", currentline)
              else
			    if (CharDump.GlobalSettings.AllSkills) then
  			      currentline = AddLine(championskillname .. ": not learned", currentline)
				end
              end			  
			else
			  -- passive
			  if (unlocklevel <= constellationpoints) then
  			    currentline = AddLine(championskillname .. ": unlocked (passive)", currentline)
		      else
			    if (CharDump.GlobalSettings.AllSkills) then
    			    currentline = AddLine(championskillname .. ": unlocks at " .. unlocklevel .. " CP in Constellation (passive)", currentline)
				end
			  end
            end
		  end
		end
	  end	  
    end
	currentline = AddLine("", currentline)
	currentline = AddLine("", currentline)
	OutputData()
end

------------------------------------------------------------
-- event handler for AddOnLoaded event

local function OnAddOnLoaded(event, addonName)
  if (addonName == CharDump.Name) then
    CharDump.Settings = ZO_SavedVars:New("CharDump_Settings", CharDump.SaveFileVersion, "chardumpsettings", local_default_settings)
    CharDump.GlobalSettings = ZO_SavedVars:NewAccountWide("CharDump_Settings", CharDump.SaveFileVersion, "chardumpsettings", global_default_settings)
	CreateWindow()
	CharDump.Initialized = true
    EVENT_MANAGER:UnregisterForEvent(CharDump.Name, EVENT_ADD_ON_LOADED)
  end
  
end


------------------------------------------------------------
--- register our event handler functions

EVENT_MANAGER:RegisterForEvent(CharDump.Name, EVENT_ADD_ON_LOADED, OnAddOnLoaded)
SLASH_COMMANDS["/dumpchar"] = DumpCharacter
SLASH_COMMANDS["/dumpchar_lockwindow"]    = LockWindow
SLASH_COMMANDS["/dumpchar_unlockwindow"]  = UnlockWindowHideChat
SLASH_COMMANDS["/dumpchar_showchat"]      = ShowChat
SLASH_COMMANDS["/dumpchar_hidechat"]      = HideChat
SLASH_COMMANDS["/dumpchar_showtext"]      = ShowText
SLASH_COMMANDS["/dumpchar_hidetext"]      = HideText
SLASH_COMMANDS["/dumpchar_writefile"]     = WriteFile
SLASH_COMMANDS["/dumpchar_nofile"]        = NoFile
SLASH_COMMANDS["/dumpchar_allskills"]     = AllSkills
SLASH_COMMANDS["/dumpchar_learnedskills"] = LearnedSkills
SLASH_COMMANDS["/chardump"] = DumpCharacter
SLASH_COMMANDS["/chardump_lockwindow"]    = LockWindow
SLASH_COMMANDS["/chardump_unlockwindow"]  = UnlockWindowHideChat
SLASH_COMMANDS["/chardump_showchat"]      = ShowChat
SLASH_COMMANDS["/chardump_hidechat"]      = HideChat
SLASH_COMMANDS["/chardump_showtext"]      = ShowText
SLASH_COMMANDS["/chardump_hidetext"]      = HideText
SLASH_COMMANDS["/chardump_writefile"]     = WriteFile
SLASH_COMMANDS["/chardump_nofile"]        = NoFile
SLASH_COMMANDS["/chardump_allskills"]     = AllSkills
SLASH_COMMANDS["/chardump_learnedskills"] = LearnedSkills

